@@ -46,12 +46,13 @@ class VSD6Parser : public VSDParser
 public:
   explicit VSD6Parser(WPXInputStream *input, libwpg::WPGPaintInterface *painter);
   ~VSD6Parser();
+protected:
+  virtual bool getChunkHeader(WPXInputStream *input);
 private:
-  bool getChunkHeader(WPXInputStream *input);
   void readText(WPXInputStream *input);
-  void readCharIX(WPXInputStream *input);
-  void readParaIX(WPXInputStream *input);
-  void readFillAndShadow(WPXInputStream *input);
+  virtual void readCharIX(WPXInputStream *input);
+  virtual void readParaIX(WPXInputStream *input);
+  virtual void readFillAndShadow(WPXInputStream *input);
   void readName(WPXInputStream *input);
   void readTextField(WPXInputStream *input);
 
