@@ -53,6 +53,62 @@ libvisio::VSDParser::~VSDParser()
 {
 }
 
+bool libvisio::VSDParser::getChunkHeader(WPXInputStream *input)
+{
+  unsigned char tmpChar = 0;
+  while (!input->atEOS() && !tmpChar)
+    tmpChar = readU8(input);
+
+  if (input->atEOS())
+    return false;
+  else
+    input->seek(-1, WPX_SEEK_CUR);
+
+  m_header.chunkType = readU32(input);
+  m_header.id = readU32(input);
+  m_header.list = readU32(input);
+
+  // Certain chunk types seem to always have a trailer
+  m_header.trailer = 0;
+  if (m_header.list != 0 || m_header.chunkType == 0x71 || m_header.chunkType == 0x70 ||
+      m_header.chunkType == 0x6b || m_header.chunkType == 0x6a || m_header.chunkType == 0x69 ||
+      m_header.chunkType == 0x66 || m_header.chunkType == 0x65 || m_header.chunkType == 0x2c)
+    m_header.trailer += 8; // 8 byte trailer
+
+  m_header.dataLength = readU32(input);
+  m_header.level = readU16(input);
+  m_header.unknown = readU8(input);
+
+  unsigned trailerChunks [14] = {0x64, 0x65, 0x66, 0x69, 0x6a, 0x6b, 0x6f, 0x71,
+                                 0x92, 0xa9, 0xb4, 0xb6, 0xb9, 0xc7
+                                };
+  // Add word separator under certain circumstances for v11
+  // Below are known conditions, may be more or a simpler pattern
+  if (m_header.list != 0 || (m_header.level == 2 && m_header.unknown == 0x55) ||
+      (m_header.level == 2 && m_header.unknown == 0x54 && m_header.chunkType == 0xaa)
+      || (m_header.level == 3 && m_header.unknown != 0x50 && m_header.unknown != 0x54))
+  {
+    m_header.trailer += 4;
+  }
+
+  for (unsigned i = 0; i < 14; i++)
+  {
+    if (m_header.chunkType == trailerChunks[i] && m_header.trailer != 12 && m_header.trailer != 4)
+    {
+      m_header.trailer += 4;
+      break;
+    }
+  }
+
+  // Some chunks never have a trailer
+  if (m_header.chunkType == 0x1f || m_header.chunkType == 0xc9 ||
+      m_header.chunkType == 0x2d || m_header.chunkType == 0xd1)
+  {
+    m_header.trailer = 0;
+  }
+  return true;
+}
+
 bool libvisio::VSDParser::parseMain()
 {
   if (!m_input)
@@ -62,14 +118,15 @@ bool libvisio::VSDParser::parseMain()
   // Seek to trailer stream pointer
   m_input->seek(0x24, WPX_SEEK_SET);
 
-  m_input->seek(8, WPX_SEEK_CUR);
-  unsigned offset = readU32(m_input);
-  unsigned length = readU32(m_input);
-  unsigned short format = readU16(m_input);
-  bool compressed = ((format & 2) == 2);
+  Pointer trailerPointer;
+  readPointer(m_input, trailerPointer);
+  bool compressed = ((trailerPointer.Format & 2) == 2);
+  unsigned shift = 0;
+  if (compressed)
+    shift = 4;
 
-  m_input->seek(offset, WPX_SEEK_SET);
-  VSDInternalStream trailerStream(m_input, length, compressed);
+  m_input->seek(trailerPointer.Offset, WPX_SEEK_SET);
+  VSDInternalStream trailerStream(m_input, trailerPointer.Length, compressed);
 
   std::vector<std::map<unsigned, XForm> > groupXFormsSequence;
   std::vector<std::map<unsigned, unsigned> > groupMembershipsSequence;
@@ -78,7 +135,7 @@ bool libvisio::VSDParser::parseMain()
   VSDStylesCollector stylesCollector(groupXFormsSequence, groupMembershipsSequence, documentPageShapeOrders);
   m_collector = &stylesCollector;
   VSD_DEBUG_MSG(("VSDParser::parseMain 1st pass\n"));
-  if (!parseDocument(&trailerStream))
+  if (!parseDocument(&trailerStream, shift))
     return false;
 
   _handleLevelChange(0);
@@ -88,17 +145,17 @@ bool libvisio::VSDParser::parseMain()
   VSDContentCollector contentCollector(m_painter, groupXFormsSequence, groupMembershipsSequence, documentPageShapeOrders, styles, m_stencils);
   m_collector = &contentCollector;
   VSD_DEBUG_MSG(("VSDParser::parseMain 2nd pass\n"));
-  if (!parseDocument(&trailerStream))
+  if (!parseDocument(&trailerStream, shift))
     return false;
 
   return true;
 }
 
-bool libvisio::VSDParser::parseDocument(WPXInputStream *input)
+bool libvisio::VSDParser::parseDocument(WPXInputStream *input, unsigned shift)
 {
   try
   {
-    handleStreams(input, 4, 0);
+    handleStreams(input, VSD_TRAILER_STREAM, shift, 0);
     return true;
   }
   catch (...)
@@ -113,36 +170,51 @@ bool libvisio::VSDParser::extractStencils()
   return parseMain();
 }
 
-void libvisio::VSDParser::handleStreams(WPXInputStream *input, unsigned shift, unsigned level)
+void libvisio::VSDParser::readPointer(WPXInputStream *input, Pointer &ptr)
 {
+  ptr.Type = readU32(input);
+  input->seek(4, WPX_SEEK_CUR); // Skip dword
+  ptr.Offset = readU32(input);
+  ptr.Length = readU32(input);
+  ptr.Format = readU16(input);
+}
+
+void libvisio::VSDParser::readPointerInfo(WPXInputStream *input, unsigned /* ptrType */, unsigned shift, unsigned &listSize, int &pointerCount)
+{
+  VSD_DEBUG_MSG(("VSDParser::readPointerInfo\n"));
+  input->seek(shift, WPX_SEEK_SET);
+  unsigned offset = readU32(input);
+  input->seek(offset+shift-4, WPX_SEEK_SET);
+  listSize = readU32(input);
+  pointerCount = readS32(input);
+  input->seek(4, WPX_SEEK_CUR);
+}
+
+void libvisio::VSDParser::handleStreams(WPXInputStream *input, unsigned ptrType, unsigned shift, unsigned level)
+{
+  VSD_DEBUG_MSG(("VSDParser::HandleStreams\n"));
   std::vector<unsigned> pointerOrder;
   std::map<unsigned, libvisio::Pointer> PtrList;
   std::map<unsigned, libvisio::Pointer> FontFaces;
-  unsigned i = 0;
 
   try
   {
     // Parse out pointers to streams
-    input->seek(shift, WPX_SEEK_SET);
-    unsigned offset = readU32(input);
-    input->seek(offset+shift-4, WPX_SEEK_SET);
-    unsigned listSize = readU32(input);
-    unsigned pointerCount = readU32(input);
-    input->seek(4, WPX_SEEK_CUR);
-    for (i = 0; i < pointerCount; i++)
+    unsigned listSize = 0;
+    int pointerCount = 0;
+    readPointerInfo(input, ptrType, shift, listSize, pointerCount);
+    for (int i = 0; i < pointerCount; i++)
     {
       Pointer ptr;
-      ptr.Type = readU32(input);
-      input->seek(4, WPX_SEEK_CUR); // Skip dword
-      ptr.Offset = readU32(input);
-      ptr.Length = readU32(input);
-      ptr.Format = readU16(input);
+      readPointer(input, ptr);
       if (ptr.Type == VSD_FONTFACES)
         FontFaces[i] = ptr;
       else if (ptr.Type != 0)
         PtrList[i] = ptr;
     }
-    for (i = 0; i < listSize; ++i)
+    if (listSize <= 1)
+      listSize = 0;
+    while (listSize--)
       pointerOrder.push_back(readU32(input));
   }
   catch (const EndOfStreamException &)
@@ -158,9 +230,9 @@ void libvisio::VSDParser::handleStreams(WPXInputStream *input, unsigned shift, u
 
   if (!pointerOrder.empty())
   {
-    for (i=0; i < pointerOrder.size(); ++i)
+    for (unsigned j=0; j < pointerOrder.size(); ++j)
     {
-      iter = PtrList.find(pointerOrder[i]);
+      iter = PtrList.find(pointerOrder[j]);
       if (iter != PtrList.end())
       {
         handleStream(iter->second, iter->first, level+1);
@@ -173,9 +245,9 @@ void libvisio::VSDParser::handleStreams(WPXInputStream *input, unsigned shift, u
 
 }
 
-
 void libvisio::VSDParser::handleStream(const Pointer &ptr, unsigned idx, unsigned level)
 {
+  VSD_DEBUG_MSG(("VSDParser::HandleStream %u type 0x%x\n", idx, ptr.Type));
   m_header.level = level;
   m_header.id = idx;
   m_header.chunkType = ptr.Type;
@@ -184,11 +256,8 @@ void libvisio::VSDParser::handleStream(const Pointer &ptr, unsigned idx, unsigne
   bool compressed = ((ptr.Format & 2) == 2);
   m_input->seek(ptr.Offset, WPX_SEEK_SET);
   VSDInternalStream tmpInput(m_input, ptr.Length, compressed);
+  m_header.dataLength = tmpInput.getSize();
   unsigned shift = compressed ? 4 : 0;
-
-  VSD_DEBUG_MSG(("VSDParser::handleStream: level %i, ptr.Type 0x%.8x, ptr.Offset 0x%.8x, ptr.Length 0x%.8x, ptr.Format 0x%.4x\n",
-                 level, ptr.Type, ptr.Offset, ptr.Length, ptr.Format));
-
   switch (ptr.Type)
   {
   case VSD_STYLES:
@@ -201,7 +270,7 @@ void libvisio::VSDParser::handleStream(const Pointer &ptr, unsigned idx, unsigne
   case VSD_PAGE:
     if (m_extractStencils)
       return;
-    if (ptr.Format == 0xd2 || ptr.Format == 0xd6)
+    if (!(ptr.Format&0x1))
       m_isBackgroundPage = true;
     else
       m_isBackgroundPage = false;
@@ -243,9 +312,9 @@ void libvisio::VSDParser::handleStream(const Pointer &ptr, unsigned idx, unsigne
     if (ptr.Length > 4)
       handleBlob(&tmpInput, level+1);
     if ((ptr.Format >> 4) == 0x5 && ptr.Type != VSD_COLORS)
-      handleStreams(&tmpInput, shift, level+1);
+      handleStreams(&tmpInput, ptr.Type, shift, level+1);
   }
-  else if ((ptr.Format >> 4) == 0xd || (ptr.Format >> 4) == 0x8)
+  else if ((ptr.Format >> 4) == 0xd || (ptr.Format >> 4) == 0xc || (ptr.Format >> 4) == 0x8)
     handleChunks(&tmpInput, level+1);
 
   switch (ptr.Type)
@@ -311,7 +380,6 @@ void libvisio::VSDParser::handleBlob(WPXInputStream *input, unsigned level)
   }
 }
 
-
 void libvisio::VSDParser::handleChunks(WPXInputStream *input, unsigned level)
 {
   long endPos = 0;
@@ -323,7 +391,7 @@ void libvisio::VSDParser::handleChunks(WPXInputStream *input, unsigned level)
     endPos = m_header.dataLength+m_header.trailer+input->tell();
 
     _handleLevelChange(m_header.level);
-    VSD_DEBUG_MSG(("Shape: parsing chunk type 0x%x\n", m_header.chunkType));
+    VSD_DEBUG_MSG(("VSDParser::handleChunks - parsing chunk type 0x%x\n", m_header.chunkType));
     handleChunk(input);
     input->seek(endPos, WPX_SEEK_SET);
   }
@@ -354,6 +422,9 @@ void libvisio::VSDParser::handleChunk(WPXInputStream *input)
   case VSD_LINE:
     readLine(input);
     break;
+  case VSD_PROP_LIST:
+    readPropList(input);
+    break;
   case VSD_FILL_AND_SHADOW:
     readFillAndShadow(input);
     break;
@@ -740,10 +811,14 @@ void libvisio::VSDParser::readParaList(WPXInputStream *input)
     m_collector->collectUnhandledChunk(m_header.id, m_header.level);
 }
 
+void libvisio::VSDParser::readPropList(WPXInputStream * /* input */)
+{
+}
+
 void libvisio::VSDParser::readPage(WPXInputStream *input)
 {
   input->seek(8, WPX_SEEK_CUR); //sub header length and children list length
-  uint32_t backgroundPageID = readU32(input);
+  unsigned backgroundPageID = readU32(input);
   m_collector->collectPage(m_header.id, m_header.level, backgroundPageID, m_isBackgroundPage);
 }
 
@@ -805,8 +880,8 @@ void libvisio::VSDParser::readXFormData(WPXInputStream *input)
   m_shape.m_xform.pinLocY = readDouble(input);
   input->seek(1, WPX_SEEK_CUR);
   m_shape.m_xform.angle = readDouble(input);
-  m_shape.m_xform.flipX = (readU8(input) != 0);
-  m_shape.m_xform.flipY = (readU8(input) != 0);
+  m_shape.m_xform.flipX = !!readU8(input);
+  m_shape.m_xform.flipY = !!readU8(input);
 }
 
 void libvisio::VSDParser::readTxtXForm(WPXInputStream *input)
@@ -833,9 +908,9 @@ void libvisio::VSDParser::readTxtXForm(WPXInputStream *input)
 void libvisio::VSDParser::readShapeId(WPXInputStream *input)
 {
   if (!m_isShapeStarted)
-    m_shapeList.addShapeId(m_header.id, readU32(input));
+    m_shapeList.addShapeId(m_header.id, getUInt(input));
   else
-    m_shape.m_shapeList.addShapeId(m_header.id, readU32(input));
+    m_shape.m_shapeList.addShapeId(m_header.id, getUInt(input));
 }
 
 void libvisio::VSDParser::readShapeList(WPXInputStream *input)
@@ -868,7 +943,10 @@ void libvisio::VSDParser::readForeignDataType(WPXInputStream *input)
   input->seek(1, WPX_SEEK_CUR);
   double imgHeight = readDouble(input);
   unsigned foreignType = readU16(input);
-  input->seek(0xb, WPX_SEEK_CUR);
+  unsigned foreignMapMode = readU16(input);
+  if (foreignMapMode == 0x8)
+    foreignType = 0x4;
+  input->seek(0x9, WPX_SEEK_CUR);
   unsigned foreignFormat = readU32(input);
 
   if (!m_shape.m_foreign)
@@ -1338,12 +1416,12 @@ void libvisio::VSDParser::readColours(WPXInputStream *input)
 {
   input->seek(6, WPX_SEEK_SET);
   unsigned numColours = readU8(input);
-  Colour tmpColour;
   input->seek(1, WPX_SEEK_CUR);
   m_colours.clear();
 
   for (unsigned i = 0; i < numColours; i++)
   {
+    Colour tmpColour;
     tmpColour.r = readU8(input);
     tmpColour.g = readU8(input);
     tmpColour.b = readU8(input);
@@ -1372,12 +1450,14 @@ void libvisio::VSDParser::readFont(WPXInputStream *input)
 
 void libvisio::VSDParser::readFontIX(WPXInputStream *input)
 {
+  long tmpAdjust = input->tell();
   input->seek(2, WPX_SEEK_CUR);
-  unsigned char codePage = readU8(input);
-  input->seek(3, WPX_SEEK_CUR);
+  unsigned char codePage = (unsigned char)(getUInt(input) & 0xff);
+  tmpAdjust -= input->tell();
+
   ::WPXBinaryData textStream;
 
-  for (unsigned i = 0; i < m_header.dataLength - 6; i++)
+  for (int i = 0; i < m_header.dataLength + tmpAdjust; i++)
   {
     unsigned char curchar = readU8(input);
     if (curchar == 0)
@@ -1443,6 +1523,251 @@ void libvisio::VSDParser::readPageSheet(WPXInputStream * /* input */)
   m_collector->collectPageSheet(m_header.id, m_header.level);
 }
 
+void libvisio::VSDParser::readText(WPXInputStream *input)
+{
+  input->seek(8, WPX_SEEK_CUR);
+  ::WPXBinaryData textStream;
+
+  // Read up to end of chunk in byte pairs (except from last 2 bytes)
+  unsigned long numBytesRead = 0;
+  const unsigned char *tmpBuffer = input->read(m_header.dataLength - 8, numBytesRead);
+  if (numBytesRead)
+  {
+    if (m_isStencilStarted)
+    {
+      VSD_DEBUG_MSG(("Found stencil text\n"));
+    }
+    textStream.append(tmpBuffer, numBytesRead);
+    m_shape.m_text = textStream;
+    m_shape.m_textFormat = libvisio::VSD_TEXT_UTF16;
+  }
+
+}
+
+void libvisio::VSDParser::readCharIX(WPXInputStream *input)
+{
+  VSDFont fontFace;
+  unsigned charCount = readU32(input);
+  unsigned fontID = readU16(input);
+  VSDName font;
+  std::map<unsigned, VSDName>::const_iterator iter = m_fonts.find(fontID);
+  if (iter != m_fonts.end())
+    font = iter->second;
+  input->seek(1, WPX_SEEK_CUR);  // Color ID
+  Colour fontColour;            // Font Colour
+  fontColour.r = readU8(input);
+  fontColour.g = readU8(input);
+  fontColour.b = readU8(input);
+  fontColour.a = readU8(input);
+
+  bool bold(false);
+  bool italic(false);
+  bool underline(false);
+  bool doubleunderline(false);
+  bool strikeout(false);
+  bool doublestrikeout(false);
+  bool allcaps(false);
+  bool initcaps(false);
+  bool smallcaps(false);
+  bool superscript(false);
+  bool subscript(false);
+  unsigned char fontMod = readU8(input);
+  if (fontMod & 1) bold = true;
+  if (fontMod & 2) italic = true;
+  if (fontMod & 4) underline = true;
+  if (fontMod & 8) smallcaps = true;
+  fontMod = readU8(input);
+  if (fontMod & 1) allcaps = true;
+  if (fontMod & 2) initcaps = true;
+  fontMod = readU8(input);
+  if (fontMod & 1) superscript = true;
+  if (fontMod & 2) subscript = true;
+
+  input->seek(4, WPX_SEEK_CUR);
+  double fontSize = readDouble(input);
+
+  fontMod = readU8(input);
+  if (fontMod & 1) doubleunderline = true;
+  if (fontMod & 4) strikeout = true;
+  if (fontMod & 0x20) doublestrikeout = true;
+
+  if (m_isInStyles)
+    m_collector->collectCharIXStyle(m_header.id, m_header.level, charCount, font, fontColour, fontSize,
+                                    bold, italic, underline, doubleunderline, strikeout, doublestrikeout,
+                                    allcaps, initcaps, smallcaps, superscript, subscript);
+  else
+  {
+    if (m_isStencilStarted)
+    {
+      VSD_DEBUG_MSG(("Found stencil character style\n"));
+    }
+
+    m_shape.m_charStyle.override(VSDOptionalCharStyle(charCount, font, fontColour, fontSize,
+                                 bold, italic, underline, doubleunderline, strikeout, doublestrikeout,
+                                 allcaps, initcaps, smallcaps, superscript, subscript));
+    m_shape.m_charList.addCharIX(m_header.id, m_header.level, charCount, font, fontColour, fontSize,
+                                 bold, italic, underline, doubleunderline, strikeout, doublestrikeout,
+                                 allcaps, initcaps, smallcaps, superscript, subscript);
+  }
+}
+
+void libvisio::VSDParser::readParaIX(WPXInputStream *input)
+{
+  unsigned charCount = readU32(input);
+  input->seek(1, WPX_SEEK_CUR);
+  double indFirst = readDouble(input);
+  input->seek(1, WPX_SEEK_CUR);
+  double indLeft = readDouble(input);
+  input->seek(1, WPX_SEEK_CUR);
+  double indRight = readDouble(input);
+  input->seek(1, WPX_SEEK_CUR);
+  double spLine = readDouble(input);
+  input->seek(1, WPX_SEEK_CUR);
+  double spBefore = readDouble(input);
+  input->seek(1, WPX_SEEK_CUR);
+  double spAfter = readDouble(input);
+  unsigned char align = readU8(input);
+  input->seek(26, WPX_SEEK_CUR);
+  unsigned flags = readU32(input);
+
+  if (m_isInStyles)
+    m_collector->collectParaIXStyle(m_header.id, m_header.level, charCount, indFirst, indLeft, indRight,
+                                    spLine, spBefore, spAfter, align, flags);
+  else
+  {
+    if (m_isStencilStarted)
+    {
+      VSD_DEBUG_MSG(("Found stencil paragraph style\n"));
+    }
+
+    m_shape.m_paraStyle.override(VSDOptionalParaStyle(charCount, indFirst, indLeft, indRight,
+                                 spLine, spBefore, spAfter, align, flags));
+    m_shape.m_paraList.addParaIX(m_header.id, m_header.level, charCount, indFirst, indLeft, indRight,
+                                 spLine, spBefore, spAfter, align, flags);
+  }
+}
+
+
+void libvisio::VSDParser::readFillAndShadow(WPXInputStream *input)
+{
+  Colour colourFG = _colourFromIndex(readU8(input));
+  input->seek(3, WPX_SEEK_CUR);
+  double fillFGTransparency = (double)readU8(input) / 255.0;
+  Colour colourBG = _colourFromIndex(readU8(input));
+  input->seek(3, WPX_SEEK_CUR);
+  double fillBGTransparency = (double)readU8(input) / 255.0;
+  unsigned char fillPattern = readU8(input);
+  input->seek(1, WPX_SEEK_CUR);
+  Colour shfgc;            // Shadow Foreground Colour
+  shfgc.r = readU8(input);
+  shfgc.g = readU8(input);
+  shfgc.b = readU8(input);
+  shfgc.a = readU8(input);
+  input->seek(5, WPX_SEEK_CUR); // Shadow Background Colour skipped
+  unsigned char shadowPattern = readU8(input);
+// only version 11 after that point
+  input->seek(2, WPX_SEEK_CUR); // Shadow Type and Value format byte
+  double shadowOffsetX = readDouble(input);
+  input->seek(1, WPX_SEEK_CUR); // Value format byte
+  double shadowOffsetY = readDouble(input);
+
+
+
+  if (m_isInStyles)
+    m_collector->collectFillStyle(m_header.level, colourFG, colourBG, fillPattern,
+                                  fillFGTransparency, fillBGTransparency, shadowPattern, shfgc,
+                                  shadowOffsetX, shadowOffsetY);
+  else
+  {
+    if (m_isStencilStarted)
+    {
+      VSD_DEBUG_MSG(("Found stencil fill\n"));
+    }
+    m_shape.m_fillStyle.override(VSDOptionalFillStyle(colourFG, colourBG, fillPattern, fillFGTransparency,
+                                 fillBGTransparency, shfgc, shadowPattern, shadowOffsetX, shadowOffsetY));
+  }
+}
+
+void libvisio::VSDParser::readName(WPXInputStream *input)
+{
+  unsigned long numBytesRead = 0;
+  const unsigned char *tmpBuffer = input->read(m_header.dataLength, numBytesRead);
+  if (numBytesRead)
+  {
+    ::WPXBinaryData name(tmpBuffer, numBytesRead);
+    m_shape.m_names[m_header.id] = VSDName(name, libvisio::VSD_TEXT_UTF16);
+  }
+}
+
+void libvisio::VSDParser::readTextField(WPXInputStream *input)
+{
+  unsigned long initialPosition = input->tell();
+  input->seek(7, WPX_SEEK_CUR);
+  unsigned char tmpCode = readU8(input);
+  if (tmpCode == 0xe8)
+  {
+    int nameId = (int)readU32(input);
+    input->seek(6, WPX_SEEK_CUR);
+    int formatStringId = (int)readU32(input);
+    m_shape.m_fields.addTextField(m_header.id, m_header.level, nameId, formatStringId);
+  }
+  else
+  {
+    double numericValue = readDouble(input);
+    input->seek(2, WPX_SEEK_CUR);
+    int formatStringId = (int)readU32(input);
+
+    unsigned blockIdx = 0;
+    unsigned length = 0;
+    unsigned short formatNumber = 0;
+    input->seek(initialPosition+0x36, WPX_SEEK_SET);
+    while (blockIdx != 2 && !input->atEOS() && (unsigned long) input->tell() < (unsigned long)(initialPosition+m_header.dataLength+m_header.trailer))
+    {
+      unsigned long inputPos = input->tell();
+      length = readU32(input);
+      if (!length)
+        break;
+      input->seek(1, WPX_SEEK_CUR);
+      blockIdx = readU8(input);
+      if (blockIdx != 2)
+        input->seek(inputPos + length, WPX_SEEK_SET);
+      else
+      {
+        input->seek(1, WPX_SEEK_CUR);
+        formatNumber = readU16(input);
+        if (0x80 != readU8(input))
+        {
+          input->seek(inputPos + length, WPX_SEEK_SET);
+          blockIdx = 0;
+        }
+        else
+        {
+          if (0xc2 != readU8(input))
+          {
+            input->seek(inputPos + length, WPX_SEEK_SET);
+            blockIdx = 0;
+          }
+          else
+            break;
+        }
+      }
+    }
+
+    if (input->atEOS())
+      return;
+
+    if (blockIdx != 2)
+    {
+      if (tmpCode == 0x28)
+        formatNumber = 200;
+      else
+        formatNumber = 0xffff;
+    }
+
+    m_shape.m_fields.addNumericField(m_header.id, m_header.level, formatNumber, numericValue, formatStringId);
+  }
+}
+
 libvisio::Colour libvisio::VSDParser::_colourFromIndex(unsigned idx)
 {
   if (idx < m_colours.size())
@@ -1450,4 +1775,14 @@ libvisio::Colour libvisio::VSDParser::_colourFromIndex(unsigned idx)
   return libvisio::Colour();
 }
 
+unsigned libvisio::VSDParser::getUInt(WPXInputStream *input)
+{
+  return readU32(input);
+}
+
+int libvisio::VSDParser::getInt(WPXInputStream *input)
+{
+  return readS32(input);
+}
+
 /* vim:set shiftwidth=2 softtabstop=2 expandtab: */
